/*
** oidentd_options.c - oidentd command-line handler.
** Copyright (C) 2001 Ryan McCabe <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
**
** $Id: oidentd_options.c,v 1.17 2001/09/14 20:59:25 odin Exp $
*/

#define _GNU_SOURCE

#include <config.h>

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <pwd.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <oidentd.h>
#include <oidentd_util.h>
#include <oidentd_inet_util.h>
#include <oidentd_user_db.h>
#include <oidentd_options.h>

#ifndef HAVE_GETOPT_LONG
#	include <missing/getopt_missing.h>
#else
#	include <getopt.h>
#endif

#ifdef MASQ_SUPPORT
#	define OPTSTRING "a:c:def:g:hmo:p:P:qr:t:u:Uv"
#else
#	define OPTSTRING "a:c:deg:ho:p:qr:t:u:Uv"
#endif

static void print_usage(void);
static void print_version(void);

static const struct option longopts[] = {
	{"address",				required_argument,	0, 'a'},
	{"charset",				required_argument,	0, 'c'},
	{"debug",				no_argument,		0, 'd'},
	{"error",				no_argument,		0, 'e'},
	{"group",				required_argument,	0, 'g'},
	{"help",				no_argument,		0, 'h'},
	{"other",				optional_argument,	0, 'o'},
	{"port",				required_argument,	0, 'p'},
	{"quiet",				no_argument,		0, 'q'},
	{"reply",				required_argument,	0, 'r'},
	{"timeout",				required_argument,	0, 't'},
	{"user",				required_argument,	0, 'u'},
#ifdef HAVE_LIBUDB
	{"udb",					no_argument,		0, 'U'},
#endif
	{"version",				no_argument,		0, 'v'},
#ifdef MASQ_SUPPORT
	{"forward",				optional_argument,	0, 'f'},
	{"masquerade",			no_argument,		0, 'm'},
	{"proxy",				required_argument,	0, 'P'},
#endif
	{NULL, 0, NULL, 0}
};

/*
** Parse any arguments passed to the program.
** Returns 0 on success, -1 on failure.
*/

int get_options(int argc, char *const argv[]) {
	int opt;
	extern char *failuser, *ret_os;
	extern u_int32_t flags;
	extern u_int32_t timeout;
	extern in_port_t listen_port;
	extern struct sockaddr_storage *addr;
	extern uid_t uid;
	extern gid_t gid;
	char *temp_os, *charset = NULL;
#ifdef MASQ_SUPPORT
	extern in_port_t fwdport;
	extern struct sockaddr_storage proxy;

	if (get_port(DEFAULT_PORT, &fwdport) == -1) {
		o_log(NORMAL, "Fatal: Bad port: \"%s\"", DEFAULT_PORT);
		return (-1);
	}
#endif

	temp_os = xstrdup("UNIX");

	if (get_port(DEFAULT_PORT, &listen_port) == -1) {
		o_log(NORMAL, "Fatal: Bad port: \"%s\"", DEFAULT_PORT);
		return (-1);
	}

	while ((opt = getopt_long(argc, argv, OPTSTRING, longopts, NULL)) != EOF) {
		switch (opt) {
			case 'a': {
				struct sockaddr_storage *temp_ss =
					xmalloc(sizeof(struct sockaddr_storage));

				if (get_addr(optarg, temp_ss) == -1) {
					o_log(NORMAL, "Fatal: Unknown host: \"%s\"", optarg);
					free(temp_ss);
					return (-1);
				}

				addr = temp_ss;
				break;
			}

			case 'c':
				if (charset != NULL)
					free(charset);
				charset = xstrdup(optarg);
				break;

			case 'd':
				flags |= DEBUG_MSGS;
				break;

			case 'e':
				flags |= HIDE_ERRORS;
				break;

#ifdef MASQ_SUPPORT
			case 'f':
			{
				char *p = optarg;

				if (p == NULL)
					p = "auth";

				if (get_port(p, &fwdport) == -1) {
					o_log(NORMAL, "Fatal: Bad port: \"%s\"", p);
					return (-1);
				}

				flags |= FORWARD;
				break;
			}

			case 'm':
				flags |= MASQ;
				break;

			case 'P':
			{
				if (get_addr(optarg, &proxy) == -1) {
					o_log(NORMAL, "Fatal: Unknown host: \"%s\"", optarg);
					return (-1);
				}

				flags |= PROXY;
				break;
			}

#endif
			case 'g':
				flags |= CHANGE_GID;
				if (find_group(optarg, &gid) == -1) {
					o_log(NORMAL, "Fatal: Unknown group: \"%s\"", optarg);
					return (-1);
				}
				break;

			case 'o':
				if (temp_os != NULL)
					free(temp_os);

				if (optarg != NULL)
					temp_os = xstrdup(optarg);
				else
					temp_os = xstrdup("OTHER");

				break;

			case 'p':
				if (get_port(optarg, &listen_port) == -1) {
					o_log(NORMAL, "Fatal: Bad port: \"%s\"", optarg);
					return (-1);
				}
				break;

			case 'q':
				flags |= QUIET;
				break;

			case 'r':
				if (failuser != NULL)
					free(failuser);
				failuser = xstrdup(optarg);
				break;

			case 't':
			{
				char *end;

				timeout = strtoul(optarg, &end, 10);
				if (*end != '\0') {
					o_log(NORMAL, "Fatal: Bad timeout value: \"%s\"", optarg);
					return (-1);
				}
				break;
			}

			case 'u':
				flags |= CHANGE_UID;
				if (find_user(optarg, &uid) == -1) {
					o_log(NORMAL, "Fatal: Unknown user: \"%s\"", optarg);
					return (-1);
				}
				break;

#ifdef HAVE_LIBUDB
			case 'U':
				flags |= USEUDB;
				break;
#endif

			case 'v':
				print_version();
				return (-1);

			case 'h':
			default:
				print_usage();
				return (-1);
		}
	}

	if (charset != NULL) {
		size_t len = strlen(temp_os) + strlen(charset) + 4;

		ret_os = xmalloc(len);
		snprintf(ret_os, len, "%s , %s", temp_os, charset);
		free(temp_os);
		free(charset);
	} else
		ret_os = temp_os;

	if (opt_enabled(DEBUG_MSGS) && opt_enabled(QUIET)) {
		o_log(NORMAL, "Fatal: The debug and quiet flags are incompatible.");
		return (-1);
	}

	/*
	** If no user was specified, use a reasonable default.
	*/

	if (!opt_enabled(CHANGE_UID)) {
		if (find_user("nobody", &uid) == -1) {
			o_log(NORMAL,
				"user \"nobody\" does not exist.  Using %u as default UID.",
				DEFAULT_UID);

			uid = DEFAULT_UID;
		}
	}

	/*
	** If no group is specified, use a reasonable default.
	*/

	if (!opt_enabled(CHANGE_GID)) {
		if (find_group("nobody", &gid) == -1) {
			if (find_group("nogroup", &gid) == -1) {
				o_log(NORMAL,
					"Groups \"nobody\" and \"nogroup\" do not exist. Using %u as default GID",
					DEFAULT_GID);

				gid = DEFAULT_GID;
			}
		}
	}

	return (0);
}

static void print_usage(void) {
	const char usage[] =
"\nUsage: oidentd [options]\n"
"-a or --address <address>    Bind to <address>\n"
"-c or --charset <charset>    Specify an alternate charset\n"
"-d or --debug                Enable debugging\n"
"-e or --error                Return \"UNKNOWN-ERROR\" for all errors\n"

#ifdef MASQ_SUPPORT
"-f or --forward [<port>]     Forward requests for masqueraded hosts to the host on port <port>\n"
"-m or --masquerade           Enable support for IP masquerading\n"
"-P or --proxy <host>         <host> acts as a proxy, forwarding connections to us\n"
#endif

"-g or --group <group>        Run with specified group or GID\n"
"-o or --other [<os>]         Return <os> instead of the operating system.  Uses \"OTHER\" if no argument is given.\n"
"-p or --port <port>          Listen for connections on specified port\n"
"-q or --quiet                Suppress normal logging\n"
"-t or --timeout <seconds>    Wait for <seconds> before closing connection\n"
"-u or --user <user>          Run as specified user or UID\n"

#ifdef HAVE_LIBUDB
"-U or --udb                  Perform lookups in UDB shared memory tables\n"
#endif

"-v or --version              Display version information and exit\n"
"-r or --reply <string>       If a query fails, pretend it succeeded, returning <string>\n"
"-h or --help                 This help message\n";

	print_version();
	printf("%s", usage);
}

static void print_version(void) {
	printf(	"oidentd " VERSION "\nRyan McCabe <odin@numb.org>\n"
			"http://dev.ojnk.net\n");
}
